/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: DocumentCache.java 532556 2007-04-26 01:18:26Z vgritsenko $
 */

package org.apache.xindice.core;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xindice.core.data.Key;
import org.apache.xindice.xml.NodeSource;
import org.apache.xindice.xml.SymbolTable;
import org.apache.xindice.xml.dom.DBDocument;
import org.apache.xindice.xml.dom.DOMParser;
import org.apache.xindice.xml.dom.DocumentImpl;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import java.util.Map;
import java.util.WeakHashMap;

/**
 * DocumentCache implements a simple Document caching system for
 * Collections.
 *
 * <small>
 * FIXME: Revisit cache implementation. Most probably, commons collections'
 *        ReferenceMap should be used instead of WeakHashMap.
 * </small>
 *
 * @version $Revision: 532556 $, $Date: 2007-04-25 21:18:26 -0400 (Wed, 25 Apr 2007) $
 */
public final class DocumentCache {

    private static final Log log = LogFactory.getLog(DocumentCache.class);

    private Map table = new WeakHashMap();

    /**
     * Obtains document from cache
     *
     * @param col document collection
     * @param key document key
     * @return document from the cache or null if not present
     */
    public Document getDocument(Collection col, Key key) {
        Object v = table.get(new CacheKey(col, key));

        if (v == null) {
            return null;

        } else if (v instanceof Document) {
            return (Document) v;

        } else if (v instanceof String) {
            try {
                Document doc = DOMParser.toDocument((String) v);
                ((DBDocument) doc).setSource(new NodeSource(col, key));
                return doc;
            } catch (Exception e) {
                if (log.isWarnEnabled()) {
                    log.warn("ignored exception", e);
                }
            }

        } else if (v instanceof byte[]) {
            try {
                SymbolTable s = col.getSymbols();
                NodeSource ns = new NodeSource(col, key);
                return new DocumentImpl((byte[]) v, s, ns);
            } catch (Exception e) {
                if (log.isWarnEnabled()) {
                    log.warn("ignored exception", e);
                }
            }
        }

        return null;
    }

    /**
     * Stores compressed document's bytes in the cache
     *
     * @param col document collection
     * @param key document key
     * @param bytes compressed document
     */
    public void putDocument(Collection col, Key key, byte[] bytes) {
        table.put(new CacheKey(col, key), bytes);
    }

    /**
     * Stores serialized document's text in the cache
     *
     * @param col document collection
     * @param key document key
     * @param chars uncompressed document
     */
    public void putDocument(Collection col, Key key, String chars) {
        table.put(new CacheKey(col, key), chars);
    }

    /**
     * Remove document from the cache
     *
     * @param col document collection
     * @param key document key
     */
    public void removeDocument(Collection col, Key key) {
        table.remove(new CacheKey(col, key));
    }

    /**
     * Obtains value of the cache control processing instruction in this document
     * @param doc document to inspect for cache control processing instruction
     * @return cache control value
     */
    public static int getCacheControl(Document doc) {
        String cache = DBDocument.CACHE;
        NodeList childNodes = doc.getChildNodes();
        int size = childNodes.getLength();
        for (int i = 0; i < size; i++) {
            Node n = childNodes.item(i);
            if (n.getNodeType() == Node.PROCESSING_INSTRUCTION_NODE && n.getNodeName().equals(DBDocument.CACHE_CONTROL)) {
                cache = n.getNodeValue().trim();
                break;
            }
        }

        if (cache != null) {
            if (cache.equals(DBDocument.CACHE)) {
                return -1;
            } else if (cache.equals(DBDocument.NOCACHE)) {
                return 0;
            } else {
                return Integer.parseInt(cache);
            }
        }

        return -1;
    }

    /**
     * CacheKey
     */
    private static class CacheKey {
        private final Collection col;
        private final Key key;
        private String strVal;

        public CacheKey(Collection col, Key key) {
            this.col = col;
            this.key = key;
        }

        public Collection getCollection() {
            return col;
        }

        public Key getKey() {
            return key;
        }

        public String toString() {
            if (strVal == null) {
                strVal = col.getCanonicalDocumentName(key);
            }
            return strVal;
        }

        public int hashCode() {
            if (strVal == null) {
                strVal = col.getCanonicalDocumentName(key);
            }
            return strVal.hashCode();
        }

        public boolean equals(Object o) {
            return (o instanceof CacheKey) && col == ((CacheKey) o).col && key.equals(((CacheKey) o).key);
        }
    }
}
