/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * $Id: SymbolTable.java 527707 2007-04-11 22:20:28Z vgritsenko $
 */

package org.apache.xindice.xml;

import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 * SymbolTable is a class that allows the Xindice Compression system to
 * resolve symbol IDs in compression/decompression passes.
 *
 * @version $Revision: 527707 $, $Date: 2007-04-11 18:20:28 -0400 (Wed, 11 Apr 2007) $
 */
public class SymbolTable implements XMLSerializable {
    static final String SYMBOLS = "symbols";
    static final String SYMBOL = "symbol";
    static final String NAME = "name";
    static final String NSURI = "nsuri";
    static final String ID = "id";

    private transient boolean dirty;
    private transient long lastModified = System.currentTimeMillis();

    private final Map symbols = new HashMap(); // String to SymbolInfo
    private final Map names = new HashMap();   // Short to SymbolInfo
    private short maxSymbol = -1;


    public SymbolTable() {
    }

    public SymbolTable(Element elem) {
        streamFromXML(elem);
    }

    public static String getLookupName(String qname, String namespaceURI) {
        StringBuffer sb = new StringBuffer(128);
        sb.append('[');
        sb.append(namespaceURI);
        sb.append(']');
        sb.append(qname);
        return sb.toString();
    }

    public static short getNormalizedSymbol(SymbolTable symbols, String lookup, NamespaceMap nsMap, boolean create) {
        if (lookup.startsWith("[")) {
            int idx = lookup.indexOf(']');
            String nsURI = lookup.substring(1, idx);
            int cidx = lookup.indexOf(idx + 1, ':');
            String name = cidx != -1 ? lookup.substring(cidx + 1) : lookup.substring(idx + 1);
            return symbols.getSymbol(getNormalizedLocalName(name, nsURI), nsURI, create);
        } else {
            int idx = lookup.indexOf(':');
            if (idx != -1) {
                String pfx = lookup.substring(0, idx);
                String nsURI = (String) nsMap.get(pfx);
                if (nsURI != null) {
                    String name = lookup.substring(idx + 1);
                    return symbols.getSymbol(getNormalizedLocalName(name, nsURI), nsURI, create);
                } else {
                    return symbols.getSymbol(lookup);
                }
            } else {
                return symbols.getSymbol(lookup, create);
            }
        }
    }

    public static String getNormalizedLocalName(String localName, String namespaceURI) {
        StringBuffer sb = new StringBuffer(32);
        sb.append("ns");
        sb.append(namespaceURI.hashCode());
        sb.append(':');
        sb.append(localName);
        return sb.toString();
    }

    public static String getNormalizedQName(String qName, String namespaceURI) {
        StringBuffer sb = new StringBuffer(32);
        sb.append("ns");
        sb.append(namespaceURI.hashCode());
        sb.append(':');
        int idx = qName.indexOf(':');
        if (idx != -1) {
            sb.append(qName.substring(idx + 1));
        } else {
            sb.append(qName);
        }
        return sb.toString();
    }

    public final boolean isDirty() {
        return dirty;
    }

    public final void setDirty(boolean dirty) {
        this.dirty = dirty;
        this.lastModified = System.currentTimeMillis();
    }

    public final SymbolInfo getSymbolInfo(String qname, String namespaceURI) {
        String lookupName = getLookupName(qname, namespaceURI);
        return (SymbolInfo) symbols.get(lookupName);
    }

    public final short getSymbol(String qname, String namespaceURI, boolean create) {
        String lookupName = getLookupName(qname, namespaceURI);

        SymbolInfo info = (SymbolInfo) symbols.get(lookupName);
        if (info != null) {
            return info.symbol;
        }

        if (create) {
            synchronized (symbols) {
                short id = ++maxSymbol;
                info = new SymbolInfo(qname, namespaceURI, id);
                symbols.put(lookupName, info);
                names.put(new Short(id), info);
                setDirty(true);
                return id;
            }
        }

        return -1;
    }

    public final short getSymbol(String qname, boolean create) {
        SymbolInfo info = (SymbolInfo) symbols.get(qname);
        if (info != null) {
            return info.symbol;
        }

        if (create) {
            synchronized (symbols) {
                short id = ++maxSymbol;
                info = new SymbolInfo(qname, id);
                symbols.put(qname, info);
                names.put(new Short(id), info);
                setDirty(true);
                return id;
            }
        }

        return -1;
    }

    public final short getSymbol(String qname, String namespaceURI) {
        return getSymbol(qname, namespaceURI, false);
    }

    public final short getSymbol(String qname) {
        return getSymbol(qname, false);
    }

    public final SymbolInfo getSymbolInfo(short symbol) {
        return (SymbolInfo) names.get(new Short(symbol));
    }

    public final String getNamespaceURI(short symbol) {
        SymbolInfo info = getSymbolInfo(symbol);
        return info != null ? info.namespaceURI : null;
    }

    public final String getName(short symbol) {
        SymbolInfo info = getSymbolInfo(symbol);
        return info != null ? info.qname : null;
    }

    public final long getLastModified() {
        return lastModified;
    }

    public final Element streamToXML(Document doc) throws DOMException {
        Element root = doc.createElement(SYMBOLS);

        synchronized (symbols) {
            Iterator list = symbols.values().iterator();
            while (list.hasNext()) {
                SymbolInfo info = (SymbolInfo) list.next();
                Element e = doc.createElement(SYMBOL);
                e.setAttribute(NAME, info.qname);
                if (info.namespaceURI != null && info.namespaceURI.length() > 0) {
                    e.setAttribute(NSURI, info.namespaceURI);
                }
                e.setAttribute(ID, Short.toString(info.symbol));
                root.appendChild(e);
            }
        }
        return root;
    }

    public final void streamFromXML(Element element) throws DOMException {
        symbols.clear();
        names.clear();
        maxSymbol = -1;

        NodeList list = element.getElementsByTagName(SYMBOL);
        int size = list.getLength();
        for (int i = 0; i < size; i++) {
            Element elem = (Element) list.item(i);

            String qname = elem.getAttribute(NAME);
            String namespaceURI = elem.getAttribute(NSURI);
            if (namespaceURI != null && namespaceURI.length() == 0) {
                namespaceURI = null;
            }

            Short id = new Short(elem.getAttribute(ID));
            if (id.shortValue() > maxSymbol) {
                maxSymbol = id.shortValue();
            }

            SymbolInfo info = new SymbolInfo(qname, namespaceURI, id.shortValue());
            if (namespaceURI != null) {
                String lookupName = getLookupName(qname, namespaceURI);
                symbols.put(lookupName, info);
            } else {
                symbols.put(qname, info);
            }

            names.put(id, info);
        }
    }

    public static final class SymbolInfo implements Serializable {
        private String namespaceURI;
        private String qname;
        private short symbol;

        public SymbolInfo(String qname, String namespaceURI, short symbol) {
            this.qname = qname;
            this.namespaceURI = namespaceURI;
            this.symbol = symbol;
        }

        public SymbolInfo(String qname, short symbol) {
            this.qname = qname;
            this.symbol = symbol;
        }

        public String getNamespaceURI() {
            return namespaceURI;
        }

        public String getQName() {
            return qname;
        }

        public short getSymbolID() {
            return symbol;
        }
    }
}
